#! /usr/bin/python3
# -*- coding: utf-8 -*-
""" VenC v0.0.0 """
""" Licence GNU/GPLv3 """

import VenC.languages as lang
import VenC.context
import yaml
import time
import cgi
import os

def replaceCharacter(string):
    i = 0
    output = str()
    for char in string:
        if ord(char) >= 32 and ord(char) <=44:
            output += '_'

        elif ord(char) >= 58 and ord(char) <= 64:
            output += '_'

        elif ord(char) >= 91 and ord(char) <= 96:
            output += '_'

        elif ord(char) >= 123:
            output += '_'

        else:
            output += char

    return output

def highLevelExportThread(keywordsDict, blogConfiguration, keywordsFilter, threadOrder, templates,filenameType):
    for keyword in keywordsDict.keys():
        ###############################################################################################################################
        # Some entries sources can be mixed up, but we first need to know what looks like directory templates when they are processed #
        ###############################################################################################################################

        paths = preprocessingPaths(keywordsDict, keyword, blogConfiguration["path"])

        ##################################################
        # We must be sure that "tag" is a dictionary key #
        ##################################################
        if keywordsFilter in paths.keys():
            entriesList = mixer(keywordsDict, keyword, paths[keywordsFilter], paths, blogConfiguration["path"])
            pages = paging(blogConfiguration["entries_per_pages"], entriesList, threadOrder)
            exportThread(pages, blogConfiguration, keywordsDict, paths, templates, keywordsFilter,filenameType)

####################################################################################################
#                                                                                                  #
#  Paths are actually computed and based on keywords                                               #
#  for further operations we need to know exactly what look like the current involved path         #
#                                                                                                  #
#  The function below return a dictionary containing need paths indexed by type of entries source  #
#                                                                                                  #
####################################################################################################

def preprocessingPaths(keywordsDict, keyword, pathTemplates):
    paths = dict()
    if keywordsDict[keyword].tagMass > 0:
        paths["tag"] = pathTemplates["tags_directory_name"].format(tag=keyword)+'/'

    if keywordsDict[keyword].dateMass > 0:
        ##########################################
        # date format sould be already processed #
        ##########################################
        paths["date"] = keyword+'/'

    if keywordsDict[keyword].authorMass > 0:
        paths["author"] = pathTemplates["authors_directory_name"].format(author=keyword)+'/'

    if keywordsDict[keyword].categoryMass > 0:
        paths["category"] = pathTemplates["categories_directory_name"].format(category=keyword.replace(' ','_').replace('\t','/')+'/')

    return paths

#########################################################################################
#                                                                                       #
#  Grab entries from multiple sources where current keyword is shared by these sources  #
#  	-keywordsDict contains all entries sorted by keyword names.                     #
#	See keyword class.                                                              #
#	-target is the current reference keyword.                                       #
#	-output is the path where processed thread will be write. It also determine	#
#	which sources will be combined.							#
#                                                                                       #
#########################################################################################

###################################################################
#  Because some source paths are the same, we want to mix source  #
###################################################################

def mixer(keywordsDict, target, output, pathList, pathTemplates):
    ref = list()
    outputList = list()
    for source in pathList.keys():
        if output == pathList[source]:
            ref.append(source)

    for source in ref:
        outputList += keywordsDict[target].relatedTo[source]

    return list(set(outputList))
  
######################################################################
#                                                                    #
#  Perform the big stuff.                                            # 
#  You shouldn't call this function but the high level method        #
#  which one take care for you of the preloading phase and the       #
#  preprocessing operations.                                         #
#                                                                    #
######################################################################

def exportThread(pages, blogConfiguration, keywordsDict, outputPaths, templates, keywordsFilter,filenameType):
    output = str()
    strAddon = '' if keywordsFilter != "rss" else "rss"
    keywordsFilter = keywordsFilter if keywordsFilter != "rss" else "root"
    i = int()
    j = int()
    columns = ['' for i in range(0, (1 if blogConfiguration["columns"] < 1 else blogConfiguration["columns"]))]
    for page in pages:
        j = 1
        navigationTool = {"pageNumber":i,"totalPages":len(pages), "filenameType": filenameType}
        if filenameType == "entry_file_name":
            navigationTool["pageNumber"] = int(page[0].split("__")[0])
        currentWrapper = VenC.context.blog(keywordsDict, blogConfiguration, pathProcessor(outputPaths[keywordsFilter]), navigationTool)

        wrapperProcessor = templateProcessor(".:",":.",'|')
        wrapperProcessor.functions["getBlog"] = currentWrapper.get
        output += wrapperProcessor.parse(templates[strAddon+"header.html"], escape=True if strAddon == "rss" else False)
        for filename in page:
            navigationTool["filenameType"] = "entry_file_name"
            currentEntry = VenC.context.entry(filename, keywordsDict, blogConfiguration, pathProcessor(outputPaths[keywordsFilter]), navigationTool)
            entryProcessor = templateProcessor(".:",":.",'|')
            entryProcessor.functions["getEntry"] = currentEntry.get
            entryProcessor.functions["getBlog"] = currentWrapper.get
            if strAddon == "rss":
                columns[0] += entryProcessor.parse(templates[strAddon+"entry.html"],escape=True)
            else:
                columns[(j-1)%blogConfiguration["columns"]] += entryProcessor.parse(templates[strAddon+"entry.html"],escape=False)
                j += 1

        if len(columns) ==1 or strAddon == "rss":
            output += columns[0]

        else:
            j = 1
            for column in columns:
                output += "<div id=\"__VENC_COLUMN_"+str(j)+"__\" class=\"__VENC_COLUMN__\">"+column+"</div>"
                j += 1                   
 
        navigationTool["filenameType"] = filenameType
        output += wrapperProcessor.parse(templates[strAddon+"footer.html"], escape=True if strAddon == "rss" else False)
        if not os.path.exists(os.getcwd()+"/blog/"+replaceCharacter(outputPaths[keywordsFilter])):
            os.makedirs(os.getcwd()+"/blog/"+replaceCharacter(outputPaths[keywordsFilter]))

        preProcessedFilename = os.getcwd()+"/blog/"+outputPaths[keywordsFilter]+blogConfiguration["path"][filenameType]
        filenameContext =	{"{page_number}": ('' if i==0 else str(i)),
				"{entry_id}": currentEntry.entry_id}

        for var in filenameContext.keys():
            preProcessedFilename = preProcessedFilename.replace(var, filenameContext[var])

        print(replaceCharacter(preProcessedFilename))
        open(replaceCharacter(preProcessedFilename),'w').write(output)
        output = str()
        columns = ['' for i in range(0, (1 if blogConfiguration["columns"] < 1 else blogConfiguration["columns"]))]
        i += 1

def paging(threadLenght, entriesList, threadOrder):
    pages = [[]]
    for entry in sorted(entriesList, reverse=threadOrder, key = lambda filename: int(filename.split("__")[0])):
        pages[-1].append(entry)
        if len(pages[-1]) == threadLenght:
            pages.append([])

    return [page for page in pages if page != list()]
 
###########################################################################
#  Quick read of templates to know which kind of thread will be exported  #
###########################################################################

def quickTemplatesProcessing(templates, functions):
    for function in functions:
        for template in templates.keys():
            if function in templates[template]:
                return True

    return False

#############################################################
#                                                           #
#  Depending on mode, getEntries returns                    #
#      -the ID for a new entry with mode = 0                #
#      -a list of file recognized as entries with mode = 1  #
#                                                           #
#############################################################

def safeStrToInt(string):
    try:
        return int(string)
    except:
        return -1

def getEntries(mode):
    try:
        entries = os.listdir(os.getcwd()+"/entries")

    except FileNotFoundError as e:
        print(lang.err[6].format(e.filename))
        exit()

    ##############################################################
    #                                                            #
    #  Filtering files.                                          #
    #  Ideally entries folder shouldn't contains anything else.  #
    #                                                            #
    ##############################################################
    
    validFilenames	= list()
    entryId 		= int()
    i			= int()
    for filename in sorted(entries, key = lambda filename: safeStrToInt(filename.split("__")[0])):
        splitFilename = filename.split("__")
        i = int()
        try:
            entryId = int(splitFilename[0])

            ##############################################################################
            #                                                                            #
            #  Check date but doesn't verify if day match with month and bissextil year  #
            #                                                                            #
            ##############################################################################

            for infoDate in splitFilename[1].split('-'):
                if i == 0 and not (0 < int(infoDate) and int(infoDate) <= 12):
                    raise ValueError

                if i == 1 and not (0 < int(infoDate) and int(infoDate) < 32):
                    raise ValueError

                if i == 2:
                    int(infoDate)

                if i == 3 and not (0 <= int(infoDate) and int(infoDate) < 24):
                    raise ValueError

                if i == 4 and not (0 <= int(infoDate) and int(infoDate) < 60):
                    raise ValueError

                i+=1

            validFilenames.append(filename)

        except ValueError:
            pass

        except IndexError:
            pass

    if mode:
        return validFilenames

    else:
        return entryId + 1

######################################################################################################################
#  Build a keywords dictionnary  where each referenced element of the dict is actually an instanciated keyword class #
######################################################################################################################

def getKeywords(getTags, getDates, getAuthors, getCategories, blogConfiguration):
    validFileNames	= getEntries(True)
    keywordsDict	= dict()
    for entry in validFileNames:
        date = time.strftime(blogConfiguration["path"]["dates_directory_name"], time.strptime(entry.split("__")[1],"%m-%d-%Y-%M-%S"))
        try:

            ###################################################################
            #                                                                 #
            #  Get the yaml part of the text file without the cariage return  #
            #                                                                 #
            ###################################################################

            stream = yaml.load(open(os.getcwd()+"/entries/"+entry, 'r').read().split("---\n")[1][:-1])
            ######################################################
            #  Depending of needs we extract only the necessary  #
            ######################################################
            if getTags:
                for tag in stream["tags"].split(','):
                    if tag != '':
                        if tag in keywordsDict.keys():
                            keywordsDict[tag].tagMass += 1
                        else:
                            keywordsDict[tag] = keyword(tagMass = 1)

                        if not entry in keywordsDict[tag].relatedTo["tag"]:
                            keywordsDict[tag].relatedTo["tag"].append(entry)
            if getDates:
                if date in keywordsDict.keys():
                    keywordsDict[date].dateMass += 1
                else:
                    keywordsDict[date] = keyword(dateMass = 1, originalDate = entry.split("__")[1])

                if not entry in keywordsDict[date].relatedTo["date"]:
                    keywordsDict[date].relatedTo["date"].append(entry)

            if getAuthors:
                for author in stream["authors"].split(','):
                    if author != '':
                        if author in keywordsDict.keys():
                            keywordsDict[author].authorMass += 1
                        else:
                            keywordsDict[author] = keyword(authorMass = 1)
                        if not entry in keywordsDict[author].relatedTo["author"]:
                            keywordsDict[author].relatedTo["author"].append(entry)
            if getCategories:
                for category in stream["categories"].split(','):
                    if category != '':
                        if category in keywordsDict.keys():
                            keywordsDict[category].categoryMass += 1
                        else:
                            keywordsDict[category] = keyword(categoryMass = 1)
                        if not entry in keywordsDict[category].relatedTo["category"]:
                            keywordsDict[category].relatedTo["category"].append(entry)

        except KeyError:
            pass

        except FileNotFoundError:
            print(lang.err[6].format(e.filename))
            exit()

    return keywordsDict

#######################################################################################################
#  Depending on given location and destination pathProcessor output the relative path between inputs  #
#######################################################################################################
 
class pathProcessor():
    def __init__(self, location):

        ###################################
        #  we do not replace "." by ".."  #
        ###################################

        self.backward = '/'.join([('.' if folder == '.' else ".." )for folder in location.split('/') if folder != ''])

    def output(self, destination):
        return self.backward+'/'+destination.replace('\t','/')

class keyword():
    def __init__(self, tagMass = 0, dateMass = 0, authorMass = 0, categoryMass = 0, originalDate = None):
        self.categoryMass	= categoryMass
        self.authorMass		= authorMass
        self.dateMass		= dateMass
        self.tagMass		= tagMass
        self.originalDate	= originalDate
        self.keywordIdentifier	= tuple()
        self.relatedTo		= {"date":list(), "tag":list(), "author":list(), "category":list()} 

    #########################################   ____________________________
    #  getMass return the sum of all masses # -( Thank you captain obvious. )
    #########################################   

    def getMass(argv):
        return str(self.tagMass + self.dateMass + self.authorMass + self.categoryMass)

    #########################################################
    #  argv[0] must be an int which define round precision  #
    #########################################################
    def getFreq(argv):
        try:
            if self.keywordIdentifier[1] == "tag":
                freq = self.tagMass/int(self.getMass)

            elif self.keywordIdentifier[1] == "date":
                freq = self.dateMass/int(self.getMass)

            elif self.keywordIdentifier[1] == "author":
                freq = self.authorMass/int(self.getMass)

            elif self.keywordIdentifier[1] == "category":
                freq = self.categoryMass/int(self.getMass)

            return str(round(freq,argv[0]))

        except:
            return str()

        return str()

#########################################################################################################################################
#                                                                                                                                       # 
#  templateProcessor is a class allowing to replace some recognized strings in given template by another strings which ones are         #
#  computed on the fly.                                                                                                                 #
#                                                                                                                                       # 
#  closeSymbol, openSymbol and separator are strings and define what we're looking for into the template. Separator allow to pass some  #
#  parameters to the function called when a markup is encounter.                                                                        #
#                                                                                                                                       # 
#  Functions are mapped into a dictionnary where the function name is associated to the function reference itself.                      #
#                                                                                                                                       # 
#########################################################################################################################################
 
class templateProcessor():
    def __init__(self, openSymbol, closeSymbol, separator, functions=dict()):
        self.closeSymbol	= closeSymbol
        self.openSymbol		= openSymbol
        self.separator		= separator
        self.functions		= functions

    def parse(self, string,escape=False):
        closeSymbolPos	= list()
        openSymbolPos	= list()
        output		= str()
        fields		= list()
        i		= int()
        while i < len(string):
            if i + len(self.openSymbol) <= len(string) and string[i:i+len(self.openSymbol)] == self.openSymbol:
                openSymbolPos.append(i)

            elif i + len(self.closeSymbol) <= len(string) and string[i:i+len(self.closeSymbol)] == self.closeSymbol:
                closeSymbolPos.append(i)

            if len(closeSymbolPos) == len(openSymbolPos) and len(closeSymbolPos) != 0 and len(openSymbolPos) != 0:
                if openSymbolPos[-1] < closeSymbolPos[0]:
                    fields = [field for field in string[openSymbolPos[-1]+2:closeSymbolPos[0]].split(self.separator) if field != '']
                    if fields[0] in self.functions.keys():
                        output = self.functions[fields[0]](fields[1:])

                    if escape:
                        return self.parse(string[:openSymbolPos[-1]]+cgi.escape(output).encode('ascii', 'xmlcharrefreplace').decode(encoding='ascii')+string[closeSymbolPos[0]+2:],escape=True)
                    else:
                        return self.parse(string[:openSymbolPos[-1]]+output+string[closeSymbolPos[0]+2:])

            i+=1
    
        return string
